# This replication script uses pre-calculated data for the irace optimization
# process
#
# If you want to re-generate this data for full replicability, please set the
# variable below to TRUE.
# NOTE: running a full-replication will take a long time even on high
# performance systems.
replication.full <- FALSE

# # LIBRARY SETUP (Uncomment as needed)
# # Imported and Suggested packages used in this replication script,
# # in their specific versions used in the paper:
# pkg.urls <-
#   c("https://cran.r-project.org/src/contrib/FNN_1.1.tar.gz",
#   "https://cran.r-project.org/src/contrib/assertthat_0.2.0.tar.gz",
#   "https://cran.r-project.org/src/contrib/smoof_1.5.1.tar.gz",
#   "https://cran.r-project.org/src/contrib/Archive/irace/irace_2.3.tar.gz",
#   "https://cran.r-project.org/src/contrib/ggplot2_2.2.1.tar.gz",
#   "https://cran.r-project.org/src/contrib/reshape2_1.4.2.tar.gz")
# for (pk in pkg.urls) install.packages(pk, repos = NULL, type = "source")
#
# # NOTE: the source files for all dependencies is also available in the folder
# # "dependency_packages", and can alternatively be installed from there.
#
# # Install MOEADr package from the source provided with this script
# # install.packages("MOEADr_1.1.0.tar.gz", repos = NULL, type = "source")

# Load packages
library("MOEADr")
library("smoof")
library("irace")
library("parallel")
library("ggplot2")

#################################
# Case Study 1: Figures 1 and 2 #
#################################

## Case Study 1.1
sphere <- function(X) {
  X.shift <- X + seq_along(X) * 0.1     # displace input parameters
  sum(X.shift ^ 2)
}

rastringin <- function(X) {
  X.shift <- X - seq_along(X) * 0.1     # displace input parameters in the opposite direction
  sum((X.shift) ^ 2 - 10 * cos(2 * pi * X.shift) + 10)
}

problem.sr <- function(X) {
  t(apply(X, MARGIN = 1,
          FUN = function(X) { c(sphere(X), rastringin(X)) }
  ))
}

problem.1 <- list(name       = "problem.sr",  # function that executes the MOP
                  xmin       = rep(-1, 30),   # minimum parameter value for each dimension
                  xmax       = rep(1, 30),    # maximum parameter value for each dimension
                  m          = 2)             # number of objectives

results <- moead(problem  = problem.1,
                 preset   = preset_moead("original"),
                 showpars = list(show.iters = "none"),
                 seed     = 42)

summary(results)
plot(results, suppress.pause = TRUE)
cat("Finished Case Study 1.1 (Figure 1).\n")

## Case Study 1.2
DTLZ2 <- make_vectorized_smoof(prob.name    = "DTLZ2",
                               dimensions   = 20,
                               n.objectives = 5)

problem.dtlz2  <- list(name = "DTLZ2",
                       xmin = rep(0, 20),
                       xmax = rep(1, 20),
                       m    = 5)

results.dtlz <- moead(problem  = problem.dtlz2,
                      preset   = preset_moead("original"),
                      decomp   = list(name = "SLD", H = 8),
                      showpars = list(show.iters = "dots"),
                      seed     = 42)

summary(results.dtlz)
plot(results.dtlz, suppress.pause = TRUE)
cat("Finished Case Study 1.2 (Figure 2).\n")

#################################
# Case Study 2: Figures 3 and 4 #
#################################
scenario                <- irace::defaultScenario()
scenario$seed           <- 42              # Seed for the experiment
scenario$targetRunner   <- "target.runner" # Runner function (def. below)
scenario$forbiddenFile  <- "forbidden.txt" # forbidden configs
scenario$debugLevel     <- 1               # minimal echoing
scenario$maxExperiments <- 1000            # Tuning budget
scenario$testNbElites   <- 7               # test all final configs
scenario$parallel       <- 1               # Run sequentially (for reproducibility)

#===============
### Build training instances
fname   <- paste0("UF_", 1:10)
dims    <- c(10:19,
             21:30)

allfuns            <- expand.grid(fname, dims, stringsAsFactors = FALSE)
scenario$instances <- paste0(allfuns[,1], "_", allfuns[,2])

for (i in 1:nrow(allfuns)){
  assign(x     = scenario$instances[i],
         value = make_vectorized_smoof(prob.name  = "UF",
                                       dimensions = allfuns[i, 2],
                                       id         = as.numeric(strsplit(allfuns[i, 1], "_") [[1]][2]))) }

### Build test instances
test.dims              <- 20
allfuns                <- expand.grid(fname, test.dims, stringsAsFactors = FALSE)
scenario$testInstances <- paste0(allfuns[,1], "_", allfuns[,2])

for (i in 1:nrow(allfuns)){
  assign(x     = scenario$testInstances[i],
         value = make_vectorized_smoof(prob.name  = "UF",
                                       dimensions = allfuns[i, 2],
                                       id         = as.numeric(strsplit(allfuns[i, 1], "_") [[1]][2]))) }

# Read tunable parameter list from file
parameters <- readParameters("parameters.txt")

target.runner <- function(experiment, scenario){
  force(experiment)
  force(scenario)
  conf <- experiment$configuration
  inst <- experiment$instance

  # Assemble moead input lists
  ## 1. Problem
  fdef    <- unlist(strsplit(inst, split = "_"))
  uffun   <- smoof::makeUFFunction(dimensions = as.numeric(fdef[3]),
                                   id         = as.numeric(fdef[2]))
  fattr   <- attr(uffun, "par.set")
  problem <- list(name       = inst,
                  xmin       = fattr$pars$x$lower,
                  xmax       = fattr$pars$x$upper,
                  m          = attr(uffun, "n.objectives"))

  ## 2. Decomp (fixed)
  decomp <- list(name = "Uniform")
  if (problem$m == 2){ # <-- 2 objectives
    decomp$N <- 100
  } else { # <-- 3 objectives
    decomp$N <- 150
  }

  ## 3. Neighbors (3 top parameters)
  neighbors <- list(name    = conf$neighbor.name,
                    T       = conf$T,
                    delta.p = conf$delta.p)

  ## 4. Aggfun (fixed)
  aggfun <- list(name = "AWT")

  ## 5. Update (1 top parameter)
  update <- list(name       = "restricted",
                 nr         = conf$nr,
                 UseArchive = FALSE)

  ## 6. Scaling (fixed)
  scaling <- list(name = "simple")

  ## 7. Constraint handling (fixed)
  constraint<- list(name = "none")

  ## 8. Stop criterion (fixed)
  stopcrit  <- list(list(name    = "maxeval",
                         maxeval = 50000))

  ## 9. Echoing (fixed)
  showpars  <- list(show.iters = "none")

  ## 10. Variation stack (3 top parameters)
  variation <- list(list(name  = "diffmut",
                         basis = conf$diffmut.basis),
                    list(name  = "binrec",
                         rho   = conf$binrec.rho),
                    list(name  = "polymut",
                         pm    = 1 / length(problem$xmin),
                         etam  = conf$polymut.etam),
                    list(name = "truncate"))

  # Run MOEA/D
  out <- moead(preset     = NULL,
               problem    = problem,    decomp    = decomp,
               aggfun     = aggfun,     neighbors = neighbors,
               variation  = variation,  update    = update,
               constraint = constraint, scaling   = scaling,
               stopcrit   = stopcrit,   showpars  = showpars)

  # return IGD based on reference data
  Yref <- as.matrix(read.table(file.path("pf_data",
                                         paste0(fdef[1], fdef[2], ".dat"))))
  return(list(cost = calcIGD(Y = out$Y, Yref = Yref)))
}

## Running the experiment
if (replication.full == TRUE) {
  os.type <- sessionInfo()$running
  if (os.type != "Ubuntu 16.04.3 LTS") {
    cat("ATTENTION: Detected R running under", os.type)
    cat("\nThis experiment (Case Study 2) was run under Ubuntu 16.04.3 LTS")
    cat("\nResults may be different from those reported on the paper when the")
    cat("\nexperiment is run on a different OS.")
    if (interactive()) {
        invisible(readline("[Press ENTER to continue]"))
    }
  }

  irace.output <- irace::irace(scenario, parameters)
  saveRDS(irace.output, "RESULTS.rds")
  file.copy(from = "irace.Rdata", to = "irace-tuning.Rdata",
            overwrite = TRUE)


  ## Test returned configurations on test instances
  testing.main(logFile = "irace-tuning.Rdata")
  file.copy(from = "irace.Rdata", to = "irace-testing.Rdata",
            overwrite = TRUE)
}

### Loading results
load("irace-testing.Rdata")

### Precondition results
results <- iraceResults$testing$experiments
results <- cbind(data.frame(Problem    = allfuns[, 1],
                            Objectives = rep(c(rep(2, 7),
                                               rep(3, 3)),
                                             times = 3),
                            stringsAsFactors = FALSE),
                 results)
results$Problem <- formatC(as.numeric(do.call(rbind,
                                              strsplit(results$Problem,
                                                       split = "_"))[,2]),
                           width = 2,
                           format = "d",
                           flag = "0")
res.df2 <- reshape2::melt(results,
                          id.vars = c("Problem", "Objectives"),
                          value.name = "IGD")
names(res.df2)[3] <- "Configuration"

## Plot resulting IGD of best configurations returned, by problem and dimension
ml2 <- ggplot2::theme(axis.title  = ggplot2::element_text(size = 18),
                      axis.text   = ggplot2::element_text(size = 10),
                      legend.text = ggplot2::element_text(size = 18))

mp <- ggplot2::ggplot(res.df2, ggplot2::aes(x      = Problem,
                                            y      = IGD,
                                            colour = Configuration,
                                            group  = Configuration))
plot(mp +
          ggplot2::geom_boxplot(ggplot2::aes(fill   = NULL,
                                             colour = NULL,
                                             group  = Problem),
                                alpha = 0.6,
                                lwd   = 0.1) +
          ggplot2::geom_point() + ggplot2::geom_line() + ml2)

## Plot relevant parameter frequency:
### Select output "elite" variables
indx <- iraceResults$allElites[[length(iraceResults$allElites)]]
finalConfs <- iraceResults$allConfigurations[indx,]

### change internal structure of "parameters" to allow using function
### irace::parameterFrequency() to generate final plots
mypars <- iraceResults$parameters
mypars$names      <- c("T", "delta.p", "polymut.etam", "binrec.rho")
mypars$nbVariable <- 4

### Data for table of results
cat("\n## Base data used to assemble Table 6 ##\n\n")
configurations <- getFinalElites(iraceResults = iraceResults)
print(configurations)

### Plot
irace::parameterFrequency(finalConfs, mypars, cols = 4)

cat("Finished Case Study 2 (Figures 3,4, Table 6).\n")


##########################
# Case Study 3: Figure  5#
##########################

variation_gaussmut <- function(X, mean = 0, sd = 0.1, p = 0.1, ...) {
  # You want to do some error checking on the parameters here
  # But for the sake of brevity in this case study, we are skipping it.

  R <- rnorm(length(X),    # vector of normally distributed values,
             mean = mean,  # length(R) = nrow(X) * ncol(X)
             sd   = sd)
  R <- R * (runif(length(X)) <= p) # Apply binary mask, probability = p
  return (X + R)                   # Add mutations to the solution matrix
}

ZDT1 <- make_vectorized_smoof(prob.name  = "ZDT1",
                              dimensions = 30)
problem.zdt1  <- list(name       = "ZDT1",
                      xmin       = rep(0, 30),
                      xmax       = rep(1, 30),
                      m          = 2)


myvar <- list()                                   # Initialize variation stack
myvar[[1]] <- list(name = "gaussmut", p = 0.5)    # Our new operator
myvar[[2]] <- list(name = "truncate")             # Truncation repair operator

results.orig  <- moead(problem  = problem.zdt1,
                       preset   = preset_moead("original"),
                       showpars = list(show.iters = "none"),
                       seed     = 42)
results.myvar <- moead(problem  = problem.zdt1,
                       preset   = preset_moead("original"),
                       variation = myvar,
                       showpars = list(show.iters = "none"),
                       seed     = 42)

plot(results.orig, suppress.pause = TRUE)
plot(results.myvar, suppress.pause = TRUE)

cat("Finished Case Study 3 (Figures 5a and 5b).\n")

# # Cleanup (uncomment if needed)
# remove.packages("MOEADr")
# remove.packages("ggplot2")
# remove.packages("reshape2")
# remove.packages("irace")
# remove.packages("smoof")
# remove.packages("FNN")
# remove.packages("assertthat")
